package dev.fyloz.memorycache

import io.kotest.assertions.throwables.shouldThrow
import io.kotest.core.spec.style.ShouldSpec
import io.kotest.matchers.shouldBe
import io.mockk.*

internal class BaseMemoryCacheTest : ShouldSpec({
    val cacheMap = mockk<MutableMap<Int, String>>()
    val cache = TestBaseMemoryCache(cacheMap)

    val key = Int.MAX_VALUE
    val value = "unit test value"

    fun setup_cacheMap_contains(exists: Boolean) {
        every { any() in cacheMap } returns exists
    }

    fun setup_cacheMap_get(value: String?) {
        every { cacheMap[any()] } returns value
    }

    afterEach {
        clearMocks(cacheMap)
    }

    context("contains") {
        should("return false when key not in cache") {
            // Arrange
            setup_cacheMap_contains(false)

            // Act
            val contains = key in cache

            // Assert
            contains shouldBe false
        }

        should("return true when key in cache") {
            // Arrange
            setup_cacheMap_contains(true)

            // Act
            val contains = key in cache

            // Assert
            contains shouldBe true
        }
    }

    context("get") {
        should("return value corresponding to key") {
            // Arrange
            val expectedValue = "unit test value"

            setup_cacheMap_get(expectedValue)

            // Act
            val actualValue = cache[key]

            // Assert
            actualValue shouldBe expectedValue
        }

        should("return null when no value corresponding to key") {
            // Arrange
            setup_cacheMap_get(null)

            // Act
            val actualValue = cache[key]

            // Assert
            actualValue shouldBe null
        }
    }

    context("getNotNull") {
        should("return value corresponding to key") {
            // Arrange
            setup_cacheMap_contains(true)
            setup_cacheMap_get(value)

            // Act
            val actualValue = cache.getNotNull(key)

            // Assert
            actualValue shouldBe value
        }

        should("throw NullPointerException when no value corresponding to key") {
            // Arrange
            setup_cacheMap_contains(false)
            setup_cacheMap_get(null)

            // Act
            // Assert
            shouldThrow<KeyNotFoundException> { cache.getNotNull(key) }
        }

        should("throw NullPointerException with given key when no value corresponding to key") {
            // Arrange
            setup_cacheMap_contains(false)
            setup_cacheMap_get(null)

            // Act
            val exception = shouldThrow<KeyNotFoundException> { cache.getNotNull(key) }

            // Assert
            exception.key shouldBe key
        }
    }

    context("set") {
        should("associates value to key in internal map") {
            // Arrange
            every { cacheMap[any()] = any() } just runs

            // Act
            cache[key] = value

            // Assert
            verify {
                cacheMap[key] = value
            }
            confirmVerified(cacheMap)
        }
    }

    context("remove") {
        should("removes key-value pair from internal map") {
            // Arrange
            every { cacheMap.remove(any()) } returns value

            // Act
            cache.remove(key)

            // Assert
            verify {
                cacheMap.remove(key)
            }
            confirmVerified(cacheMap)
        }
    }

    context("clear") {
        should("clears internal map") {
            // Arrange
            every { cacheMap.clear() } just runs

            // Act
            cache.clear()

            // Assert
            verify {
                cacheMap.clear()
            }
            confirmVerified(cacheMap)
        }
    }
})

private class TestBaseMemoryCache<K, V>(map: MutableMap<K, V>) : BaseMemoryCache<K, V>(map)